@@ -34,7 +34,7 @@ except ImportError:
     asyncio = None
 
 
-__version__ = '3.1'
+__version__ = '3.2'
 
 log = logging.getLogger(__name__)
 
@@ -57,6 +57,12 @@ IDENTITY_ENCODING = 'identity_encoding'
 #: for a 100-continue response from the server
 WAITING_FOR_100c = object()
 
+#: Sequence of ``(hostname, port)`` tuples that are used by
+#: `is_temp_network_error` to distinguish between permanent and temporary name
+#: resolution problems.
+DNS_TEST_HOSTNAMES=(('www.google.com', 80),
+                    ('www.iana.org', 80),
+                    ('C.root-servers.org', 53))
 
 class PollNeeded(tuple):
     '''
@@ -230,6 +236,15 @@ class ConnectionClosed(_GeneralError):
     msg = 'connection closed unexpectedly'
 
 
+class ConnectionTimedOut(_GeneralError):
+    '''
+    Raised if a regular `HTTPConnection` method (i.e., no coroutine) was
+    unable to send or receive data for the timeout specified in the
+    `HTTPConnection.timeout` attribute.
+    '''
+
+    msg = 'send/recv timeout exceeded'
+
 class _Buffer:
     '''
     This class represents a buffer with a fixed size, but varying
@@ -299,11 +314,15 @@ class _Buffer:
 
 class HTTPConnection:
     '''
-    This class encapsulates a HTTP connection. Methods whose name begin with
-    ``co_`` return coroutines. Instead of blocking, a coroutines will yield
-    a `PollNeeded` instance that encapsulates information about the IO operation
-    that would block. The coroutine should be resumed once the operation can be
-    performed without blocking.
+    This class encapsulates a HTTP connection.
+
+    Methods whose name begin with ``co_`` return coroutines. Instead of
+    blocking, a coroutines will yield a `PollNeeded` instance that encapsulates
+    information about the IO operation that would block. The coroutine should be
+    resumed once the operation can be performed without blocking.
+
+    `HTTPConnection` instances can be used as context managers. The
+    `.disconnect` method will be called on exit from the managed block.
     '''
 
     def __init__(self, hostname, port=None, ssl_context=None, proxy=None):
@@ -351,6 +370,12 @@ class HTTPConnection:
         #: Transfer encoding of the active response (if any).
         self._encoding = None
 
+        #: If a regular `HTTPConnection` method is unable to send or receive
+        #: data for more than this period (in seconds), it will raise
+        #: `ConnectionTimedOut`. Coroutines are not affected by this
+        #: attribute.
+        self.timeout = None
+
     # Implement bare-bones `io.BaseIO` interface, so that instances
     # can be wrapped in `io.TextIOWrapper` if desired.
     def writable(self):
@@ -377,7 +402,7 @@ class HTTPConnection:
         if self.proxy:
             log.debug('connecting to %s', self.proxy)
             self._sock = socket.create_connection(self.proxy)
-            eval_coroutine(self._co_tunnel())
+            eval_coroutine(self._co_tunnel(), self.timeout)
         else:
             log.debug('connecting to %s', (self.hostname, self.port))
             self._sock = socket.create_connection((self.hostname, self.port))
@@ -448,7 +473,8 @@ class HTTPConnection:
     def send_request(self, method, path, headers=None, body=None, expect100=False):
         '''placeholder, will be replaced dynamically'''
         eval_coroutine(self.co_send_request(method, path, headers=headers,
-                                            body=body, expect100=expect100))
+                                            body=body, expect100=expect100),
+                       self.timeout)
 
     def co_send_request(self, method, path, headers=None, body=None, expect100=False):
         '''Send a new HTTP request to the server
@@ -589,7 +615,7 @@ class HTTPConnection:
 
     def write(self, buf):
         '''placeholder, will be replaced dynamically'''
-        eval_coroutine(self.co_write(buf))
+        eval_coroutine(self.co_write(buf), self.timeout)
 
     def co_write(self, buf):
         '''Write request body data
@@ -614,7 +640,15 @@ class HTTPConnection:
             raise ExcessBodyData('trying to write %d bytes, but only %d bytes pending'
                                     % (len(buf), remaining))
 
-        yield from self._co_send(buf)
+        try:
+            yield from self._co_send(buf)
+        except ConnectionClosed:
+            # If the server closed the connection, we pretend that all data
+            # has been sent, so that we can still read a (buffered) error
+            # response.
+            self._out_remaining = None
+            self._pending_requests.append((method, path, None))
+            raise
 
         len_ = len(buf)
         if len_ == remaining:
@@ -636,7 +670,7 @@ class HTTPConnection:
 
     def read_response(self):
         '''placeholder, will be replaced dynamically'''
-        return eval_coroutine(self.co_read_response())
+        return eval_coroutine(self.co_read_response(), self.timeout)
 
     def co_read_response(self):
         '''Read response status line and headers
@@ -817,7 +851,7 @@ class HTTPConnection:
         '''placeholder, will be replaced dynamically'''
         if len_ is None:
             return self.readall()
-        buf = eval_coroutine(self.co_read(len_))
+        buf = eval_coroutine(self.co_read(len_), self.timeout)
 
         # Some modules like TextIOWrapper unfortunately rely on read()
         # to return bytes, and do not accept bytearrays or memoryviews.
@@ -896,7 +930,7 @@ class HTTPConnection:
 
     def readinto(self, buf):
         '''placeholder, will be replaced dynamically'''
-        return eval_coroutine(self.co_readinto(buf))
+        return eval_coroutine(self.co_readinto(buf), self.timeout)
 
     def co_readinto(self, buf):
         '''Read response body data into *buf*
@@ -1198,7 +1232,7 @@ class HTTPConnection:
 
     def readall(self):
         '''placeholder, will be replaced dynamically'''
-        return eval_coroutine(self.co_readall())
+        return eval_coroutine(self.co_readall(), self.timeout)
 
     def co_readall(self):
         '''Read and return complete response body'''
@@ -1223,7 +1257,7 @@ class HTTPConnection:
 
     def discard(self):
         '''placeholder, will be replaced dynamically'''
-        return eval_coroutine(self.co_discard())
+        return eval_coroutine(self.co_discard(), self.timeout)
 
     def co_discard(self):
         '''Read and discard current response body'''
@@ -1257,6 +1291,12 @@ class HTTPConnection:
         else:
             log.debug('already closed')
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        self.disconnect()
+        return False
 
 def _extend_HTTPConnection_docstrings():
 
@@ -1301,28 +1341,50 @@ else:
     def _join(parts):
         return b''.join(parts)
 
-def eval_coroutine(crt):
-    '''Evaluate *crt* (polling as needed) and return its result'''
+def eval_coroutine(crt, timeout=None):
+    '''Evaluate *crt* (polling as needed) and return its result
+
+    If *timeout* seconds pass without being able to send or receive
+    anything, raises `ConnectionTimedOut`.
+    '''
 
     try:
         while True:
-            assert next(crt).poll()
             log.debug('polling')
+            if not next(crt).poll(timeout=timeout):
+                raise ConnectionTimedOut()
     except StopIteration as exc:
         return exc.value
 
 def is_temp_network_error(exc):
-    '''Return true if *exc* represents a potentially temporary network problem'''
+    '''Return true if *exc* represents a potentially temporary network problem
+
+    For problems with name resolution, the exception generally does not contain
+    enough information to distinguish between an unresolvable hostname and a
+    problem with the DNS server. In this case, this function attempts to resolve
+    the addresses in `DNS_TEST_HOSTNAMES`. A name resolution problem is
+    considered permanent if at least one test hostname can be resolved, and
+    temporary if none of the test hostnames can be resolved.
+    '''
 
     if isinstance(exc, (socket.timeout, ConnectionError, TimeoutError, InterruptedError,
                         ConnectionClosed, ssl.SSLZeroReturnError, ssl.SSLEOFError,
-                        ssl.SSLSyscallError)):
+                        ssl.SSLSyscallError, ConnectionTimedOut)):
         return True
 
-    # Formally this is a permanent error. However, it may also indicate
-    # that there is currently no network connection to the DNS server
+    # The exception also unfortunately does not help us to distinguish between
+    # permanent and temporary problems. See:
+    # https://stackoverflow.com/questions/24855168/
+    # https://stackoverflow.com/questions/24855669/
     elif (isinstance(exc, (socket.gaierror, socket.herror))
           and exc.errno in (socket.EAI_AGAIN, socket.EAI_NONAME)):
+        for (hostname, port) in DNS_TEST_HOSTNAMES:
+            try:
+                socket.getaddrinfo(hostname, port)
+            except (socket.gaierror, socket.herror):
+                pass
+            else:
+                return False
         return True
 
     return False
