@@ -21,7 +21,8 @@ if __name__ == '__main__':
 
     sys.exit(pytest.main([__file__] + sys.argv[1:]))
 
-from dugong import HTTPConnection, BodyFollowing, CaseInsensitiveDict, _join
+from dugong import (HTTPConnection, BodyFollowing, CaseInsensitiveDict, _join,
+                    ConnectionClosed)
 import dugong
 from http.server import BaseHTTPRequestHandler, _quote_html
 from io import TextIOWrapper
@@ -399,12 +400,27 @@ def test_abort_read(conn):
     assert_raises(dugong.ConnectionClosed, conn.read, 200)
 
 def test_abort_co_read(conn):
-    conn.send_request('GET', '/send_3_300-byte_chunks')
-    resp = conn.read_response()
-    assert resp.status == 200
-    cofun = conn.co_read(450)
-    next(cofun)
-    conn.disconnect()
+    # We need to delay the write to ensure that we encounter a blocking read
+    delay = 10
+    while True:
+        conn.send_request('GET', '/send_3_300-byte_chunks_delay_%d_ms' % delay)
+        resp = conn.read_response()
+        assert resp.status == 200
+        cofun = conn.co_read(450)
+        try:
+            next(cofun)
+        except StopIteration:
+            # Not good, need to wait longer
+            pass
+        else:
+            break
+        finally:
+            conn.disconnect()
+
+        if delay > 5000:
+            pytest.fail('no blocking read even with %f sec sleep' % delay)
+        delay *= 2
+
     assert_raises(dugong.ConnectionClosed, next, cofun)
 
 def test_abort_write(conn):
@@ -523,6 +539,62 @@ def test_100cont_3(conn):
     conn.read_response()
     conn.readall()
 
+def test_aborted_write1(conn, monkeypatch):
+    BUFSIZE = 64*1024
+
+    # Monkeypatch request handler
+    def do_PUT(self):
+        # Read half the data, then generate error and
+        # close connection
+        self.rfile.read(BUFSIZE)
+        self.send_error(code=401, message='Please stop!')
+        self.close_connection = True
+
+    monkeypatch.setattr(MockRequestHandler, 'do_PUT', do_PUT)
+
+    # Send request
+    conn.send_request('PUT', '/big_object', body=BodyFollowing(BUFSIZE*50),
+                      expect100=True)
+    resp = conn.read_response()
+    assert resp.status == 100
+    assert resp.length == 0
+
+    # Try to write data
+    with pytest.raises(ConnectionClosed):
+        for _ in range(50):
+            conn.write(b'f' * BUFSIZE)
+
+    # Nevertheless, try to read response
+    resp = conn.read_response()
+    assert resp.status == 401
+    assert resp.reason == 'Please stop!'
+
+def test_aborted_write2(conn, monkeypatch):
+    BUFSIZE = 64*1024
+
+    # Monkeypatch request handler
+    def do_PUT(self):
+        # Read half the data, then silently close connection
+        self.rfile.read(BUFSIZE)
+        self.close_connection = True
+
+    monkeypatch.setattr(MockRequestHandler, 'do_PUT', do_PUT)
+
+    # Send request
+    conn.send_request('PUT', '/big_object', body=BodyFollowing(BUFSIZE*50),
+                      expect100=True)
+    resp = conn.read_response()
+    assert resp.status == 100
+    assert resp.length == 0
+
+    # Try to write data
+    with pytest.raises(ConnectionClosed):
+        for _ in range(50):
+            conn.write(b'f' * BUFSIZE)
+
+    # Nevertheless, try to read response
+    assert_raises(ConnectionClosed, conn.read_response)
+
 def test_tunnel(http_server):
     if http_server.use_ssl:
         pytest.skip('test does not have ssl support yet')
@@ -626,6 +698,40 @@ def test_mutable_read(conn):
     assert conn.readall() == DUMMY_DATA[pos:512]
     assert not conn.response_pending()
 
+def test_recv_timeout(conn, monkeypatch):
+    conn.timeout = 1
+
+    def do_GET(self):
+        self.send_response(200)
+        self.send_header("Content-Type", 'application/octet-stream')
+        self.send_header("Content-Length", '50')
+        self.end_headers()
+        self.wfile.write(b'x' * 25)
+        self.wfile.flush()
+    monkeypatch.setattr(MockRequestHandler, 'do_GET', do_GET)
+
+    conn.send_request('GET', '/send_something')
+    resp = conn.read_response()
+    assert resp.status == 200
+    assert conn.read(50) == b'x' * 25
+    assert_raises(dugong.ConnectionTimedOut, conn.read, 50)
+
+def test_send_timeout(conn, monkeypatch):
+    conn.timeout = 1
+
+    def do_PUT(self):
+        # Read just a tiny bit
+        self.rfile.read(256)
+    monkeypatch.setattr(MockRequestHandler, 'do_PUT', do_PUT)
+
+    # We don't know how much data can be buffered, so we
+    # claim to send a lot and do so in a loop.
+    len_ = 1024**3
+    conn.send_request('PUT', '/recv_something', body=BodyFollowing(len_))
+    with pytest.raises(dugong.ConnectionTimedOut):
+        while len_ > 0:
+            conn.write(b'x' * min(len_, 16*1024))
+
 
 DUMMY_DATA = ','.join(str(x) for x in range(10000)).encode()
 
@@ -634,8 +740,8 @@ class MockRequestHandler(BaseHTTPRequestHandler):
     server_version = "MockHTTP"
     protocol_version = 'HTTP/1.1'
 
-    #def log_message(self, format, *args):
-    #    pass
+    def log_message(self, format, *args):
+        pass
 
     def handle_expect_100(self):
         if self.handle_errors():
